"""
Base db context
"""
from typing import Optional

from sqlalchemy.sql.elements import TextClause

from ..adaptors.adaptor import DBAdaptor
from ..utils.db_functions import currentDBTimestamp
from ..utils.log import Logger


class BaseDBContext:
    """
    Base class to work with database.

    Attributes:
        logger: the current request logger
    """

    # database adaptor
    adaptor: Optional[DBAdaptor] = None
    # database config
    dbConfig: Optional["DBSetting"] = None
    # storage time - LOCAL or UTC
    storageTime: Optional[str] = None

    def __init__(self, logger: Logger):
        self.logger = logger

    @classmethod
    async def initDBContext(  # pylint: disable-msg=W0613
        cls, dbSettings: "DBSetting", storageTime: str, **kwargs
    ) -> None:
        """
        Initialize context
        Args:
            dbSettings: database settings
            storageTime: storage time
            kwargs: init db proxy kwargs
        Keyword Args:
            connectTimeout: connection timeout for postgres, 5 seconds by default
        """
        cls.dbConfig = dbSettings
        cls.storageTime = storageTime
        adaptorKwargs = {
            "dbtype": dbSettings.type,
            "dsn": dbSettings.dsn,
            "dblogin": dbSettings.userName,
            "dbpassword": dbSettings.password,
            "sessionCount": dbSettings.dbSettings.connectionPoolSize,
        }
        if "connectTimeout" in kwargs:
            adaptorKwargs["connectTimeout"] = kwargs["connectTimeout"]
        cls.adaptor: DBAdaptor = DBAdaptor(**adaptorKwargs)
        await cls.adaptor.initiateAdaptor()

    @classmethod
    async def closeDBContext(cls) -> None:
        """
        Close all database contexts
        """
        if cls.adaptor is not None:
            await cls.adaptor.close()

    @property
    def dbType(self) -> str:
        """
        Get current db type
        Raises:
            RuntimeError if db adaptor was not initialized
        """
        if self.adaptor is None:
            raise RuntimeError("Initialize DBContext first")
        return self.adaptor.dbtype

    @property
    def currentDBTimestamp(self) -> TextClause:
        """
        Get current db timestamp function
        Returns:
            function for computation current db timestamp
        """
        return currentDBTimestamp(self.dbType)
