"""
Module realizes VLException, exception for internal usages
"""
from functools import wraps
from typing import Optional

from luna3.common.exceptions import LunaApiException
from pydantic_core import PydanticCustomError
from vlutils.helpers import isCoroutineFunction

from ..pydantic.bases import isPydanticFirst
from .errors import ErrorInfo


class VLException(Exception):
    """
    VLException

    Attributes:
        error: error
        statusCode: status code for response
        isCriticalError: print trace or not
        headers: additional http headers
    """

    def __init__(
        self, error: ErrorInfo, statusCode=500, isCriticalError=True, exception=None, headers: Optional[dict] = None
    ):
        super().__init__(error.description)
        self.error = error
        self.statusCode = statusCode
        self.isCriticalError = isCriticalError
        self.exception = exception
        self.headers = headers


def exceptionWarp(error):
    """
    Decorator for catching exceptions and raise VLException with specific error.

    Args:
        error: returning error in the exception case
    Returns:
        if exception was caught, system calls error method with error
    Raises:
        VLException(error, exception=e): if catch any exception except VLException, LunaApiException
    """

    def realWarp(func):
        if isCoroutineFunction(func):

            @wraps(func)
            async def wrap(*func_args, **func_kwargs):
                try:
                    res = await func(*func_args, **func_kwargs)
                    return res
                except (VLException, LunaApiException, TypeError):
                    raise
                except Exception as exc:
                    func_args[0].logger.exception()
                    raise VLException(error, exception=exc) from exc

        else:

            @wraps(func)
            def wrap(*func_args, **func_kwargs):
                try:
                    res = func(*func_args, **func_kwargs)
                    return res
                except (VLException, LunaApiException, TypeError):
                    raise
                except Exception as exc:
                    func_args[0].logger.exception()
                    raise VLException(error, exception=exc) from exc

        return wrap

    return realWarp


class PydanticErrorInfo:
    """
    Pydantic Custom Errors constructor
    Can only be used with format() method call

    Attributes:
        errorType: error type eg 'ValueError'
        errorMsg: error message
        ctx: dict error context
    """

    __slots__ = ["errorType", "errorMsg", "ctx"]

    def __init__(self, errorType: str, errorMsg: str, ctx: dict | None = None):
        self.errorMsg = errorMsg
        self.errorType = errorType
        self.ctx = ctx

    def __call__(self, *args, **kwargs):
        return PydanticCustomError(self.errorType, self.errorMsg, self.ctx)

    def format(self, *formatArgs) -> "PydanticErrorInfo":
        """
        Formats PydanticCustomError error message.

        Args:
            formatArgs: args to format existing error message
        Returns:
            Callable copy of PydanticErrorInfo
        >>> error = PydanticErrorInfo('ValueError', 'Value is not a valid {}', {'input': 23.2})
        >>> error.format('dictionary')()
        >>> error
        Value is not a valid dictionary [type=ValueError, context={'input': 23.2}]
        """
        return PydanticErrorInfo(self.errorType, self.errorMsg.format(*formatArgs), self.ctx)

    def setContext(self, ctx: dict) -> "PydanticErrorInfo":
        """
        Set error context equal to passed dict

        Args:
            ctx: dict context
        Returns:
            Callable copy of PydanticErrorInfo

        >>> error = PydanticErrorInfo('ValueError', 'Value is not a valid dictionary')
        >>> error = error.setContext({'input': 23.2})()
        >>> error
        Value is not a valid dictionary [type=ValueError, context={'input': 23.2}]
        """
        return PydanticErrorInfo(self.errorType, self.errorMsg, ctx)


class PydanticError:
    """Pydantic custom errors"""

    PydanticValueError = PydanticErrorInfo("ValueError", "{}")
    PydanticValidationError = PydanticErrorInfo("ValidationError", "{}")
    PydanticTypeError = PydanticErrorInfo("TypeError", "{}")
    PydanticExactFieldRequiredError = PydanticErrorInfo("ValueError", "Field '{}' is required")
    PydanticStringPatternMismatchError = PydanticErrorInfo("ValueError", "String pattern mismatch - '{}'")
    PydanticFieldRequiredError = PydanticErrorInfo("ValueError", "Field required")
    PydanticInvalidDateTimeError = PydanticErrorInfo("ValueError", "Invalid datetime")
    PydanticInvalidDictError = PydanticErrorInfo("ValueError", "Input should be a valid dictionary")
    PydanticInvalidListError = PydanticErrorInfo("ValueError", "Input should be a valid list")
    PydanticInvalidStringError = PydanticErrorInfo("ValueError", "Input should be a valid string")
    PydanticValueNotANumberError = PydanticErrorInfo("TypeError", "Value is not a valid number")
    PydanticValueNotAnIntegerError = PydanticErrorInfo("TypeError", "Value is not a valid integer")
    PydanticValueNotAFloatError = PydanticErrorInfo("TypeError", "Value is not a valid float")
    PydanticUniqueItemsError = PydanticErrorInfo("ValueError", "List must have unique items only")


# TODO: remove after LUNA-6371 completed # pylint: disable-msg=fixme
def pydanticErrorWrapper(errorInfo: PydanticErrorInfo):
    """Wrapper to adapt an error for pydantic 1"""
    if isPydanticFirst:
        errorTypes = {"ValueError": ValueError, "TypeError": TypeError, "ValidationError": ValueError}
        return errorTypes[errorInfo.errorType](errorInfo.errorMsg)
    return errorInfo()
