"""
Module containt platform-specific base classes for pydantic models
"""
from datetime import datetime
from enum import Enum
from typing import Any
from uuid import UUID

import pydantic  # TODO: remove after LUNA-6371 completed # pylint: disable-msg=fixme
from vlutils.helpers import convertTimeToString
from vlutils.structures.pydantic import BaseModel as _BaseModel, replaceEqualSuffix

# simple cache for alias generator
_CACHE = {}
# TODO: remove after LUNA-6371 completed # pylint: disable-msg=fixme
isPydanticFirst = int(pydantic.__version__.split(".")[0]) < 2  # pylint: disable-msg=use-maxsplit-arg

if not isPydanticFirst:
    from pydantic import ConfigDict


def getAlias(paramName: str):
    """
    Get alias for field
    Args:
        paramName: param name in camel case
    Returns:
        param name alias
    """
    if (paramValue := _CACHE.get(paramName)) is None:
        paramValue = _CACHE[paramName] = replaceEqualSuffix(paramName)
    return paramValue


class JsonBaseModel(_BaseModel):
    """Base model for jsons"""

    # TODO: remove after LUNA-6371 completed # pylint: disable-msg=fixme
    if isPydanticFirst:

        class Config:
            """Pydantic model config."""

            extra = pydantic.Extra.forbid

            @classmethod
            def alias_generator(cls, paramName: str) -> str:  # pylint: disable-msg=C0103
                """
                Generate aliases
                Args:
                    paramName: model field name
                """
                return getAlias(paramName)

    else:
        model_config = ConfigDict(
            extra="forbid", alias_generator=getAlias
        )  # pylint: disable-msg=used-before-assignment

    @classmethod
    def _getValue(cls, value: Any) -> Any:  # pylint: disable-msg=C0103
        """Convert uuid, enum to string"""
        if isinstance(value, (UUID, Enum)):
            if (enumValue := getattr(value, "value", None)) is not None:
                return enumValue
            return str(value)
        if isinstance(value, datetime):
            return convertTimeToString(value, True)
        return value

    def asDict(self) -> dict:
        """
        Get data from initialized model
        Returns:
            dict with snake_case keys
        """
        getValue = self._getValue

        res = {}
        for fieldName, value in self.__dict__.items():
            if value is None:
                continue
            key = getAlias(fieldName)
            if hasattr(value, "asDict"):
                res[key] = value.asDict()
            elif isinstance(value, (list, set)):
                res[key] = [row.asDict() if hasattr(row, "asDict") else getValue(row) for row in value]
            else:
                res[key] = getValue(value)

        return res


class QueryBaseModel(JsonBaseModel):
    """Base model for queries"""

    # TODO: remove after LUNA-6371 completed
    if isPydanticFirst:

        class Config:
            """Pydantic model config."""

            extra = pydantic.Extra.ignore

    else:
        model_config = ConfigDict(extra="ignore")


class NullableFieldsModelMixin:
    """Mixin for models with nullable fields"""

    def asDict(self) -> dict:
        """
        Get data from initialized model
        Returns:
            dict with snake_case keys
        """
        getValue = self._getValue

        res = {}
        for fieldName, value in self.__dict__.items():
            if fieldName not in self.__fields_set__:
                continue
            key = getAlias(fieldName)
            if hasattr(value, "asDict"):
                res[key] = value.asDict()
            elif isinstance(value, (list, set)):
                res[key] = [row.asDict() if hasattr(row, "asDict") else getValue(row) for row in value]
            else:
                res[key] = getValue(value)

        return res
