"""
Module with redis sender contexts
"""
import abc
from typing import Optional

from redis import Redis

from ..utils.log import Logger
from .adaptor import RedisAdaptor


class RedisContext(abc.ABC):
    """Base class for Redis context.

    Attributes:
        channelName: A string, redis channel prefix
        logger: A Logger
        storageTime: storage time ("UTC" or "LOCAL")
        adaptor: redis adaptor, container for client
    """

    def __init__(
        self,
        host="127.0.0.1",
        port=6379,
        password: Optional[str] = None,
        sentinels: list[tuple[str, int]] | None = None,
        masterName: str | None = None,
        dbNumber: int = None,
        channelName: str = "luna-sender",
        storageTime: Optional[str] = None,
    ) -> None:
        """Initialize redis context.

        Args:
            storageTime: storage time ("UTC" or "LOCAL")
            channelName: A string, redis channel prefix
        """
        self.channelName = channelName
        self.logger = Logger("redis_db")
        self.storageTime = storageTime or "UTC"
        self.adaptor = RedisAdaptor(
            host=host, password=password, port=port, masterName=masterName, sentinels=sentinels, dbNumber=dbNumber
        )

    @property
    def redis(self) -> Redis:
        """Get redis client"""
        return self.adaptor.client

    async def close(self) -> None:
        """Close connections."""
        await self.adaptor.close()

    @abc.abstractmethod
    async def publish(self, events: list, requestId: str) -> None:
        """
        Publish events. Msg is published to channel 'luna-sender:{account_id}' in format:
        :json:object:`msg_to_luna_sender`

        Args:
            events: events
            requestId: request id
        """
